;-------------------------------------------------------------------------------
; * Mathematical Routines
; Ver 2007 Apr 27 Rev 1
;-------------------------------------------------------------------------------
; This library auto-initializes
; Note: Some of these functions are NOT backward compatible with pre-2007 code


; Sine / Cosine tables

Global Math_SinTable#[ 360 ]
Global Math_CosTable#[ 360 ]

Function Math_InitTables()
    Local Angle
    
    For Angle = 0 To 360
        Math_SinTable[ Angle ] = Sin( Angle )
        Math_CosTable[ Angle ] = Cos( Angle )
    Next 
End Function

Math_InitTables

; Wrapper functions; very fast for imprecise trigonometry

Function qSin#( Value% )
    If Value <    0 Then Return Math_SinTable[ 360 - Abs( Value ) Mod 360 ]
    If Value >= 360 Then Return Math_SinTable[ Value Mod 360 ]
    Return Math_SinTable[ Value ]
End Function 

Function qCos#( Value% )
    If Value <    0 Then Return Math_CosTable[ 360 - Abs( Value ) Mod 360 ]
    If Value >= 360 Then Return Math_CosTable[ Value Mod 360 ]
    Return Math_CosTable[ Value ]
End Function 


; Vectors / trigonometry

; polar & cartesian coordinate conversion routines
Function VectorX# ( Distance# , Angle# )    Return Sin ( Angle      ) * Distance    End Function
Function VectorY# ( Distance# , Angle# )    Return Sin ( Angle - 90 ) * Distance    End Function
Function VectorDistance# ( X# , Y# )        Return Sqr ( X * X + Y * Y )            End Function
Function VectorAngle#    ( X# , Y# )        Return -Atan2 ( X , Y ) + 180           End Function

; short hand
Function vx# ( d# , a# )    Return Sin ( a ) * d            End Function
Function vy# ( d# , a# )    Return Sin ( a - 90 ) * d       End Function
Function vd# ( x# , y# )    Return Sqr ( x * x + y * y )    End Function
Function va# ( x# , y# )    Return -ATan2 ( x , y ) + 180   End Function

; quick and imprecise
Function qvx# ( d# , a# )    Return qSin ( a ) * d            End Function
Function qvy# ( d# , a# )    Return qSin ( a - 90 ) * d       End Function


; Range limiting (overflow)

Function RangeFloat#( Value# , Max#=360 )
    If Value <    0 Then Return Max - Abs( Value ) Mod Max
    If Value >= Max Then Return Value Mod Max
    Return Value
End Function 

Function RangeInt%( Value% , Max%=360 )
    If Value <    0 Then Return Max - Abs( Value ) Mod Max
    If Value >= Max Then Return Value Mod Max
    Return Value
End Function 


; Capping (no overflow)

Function CapFloat#( Value# , Min#,Max# )
    If Value < Min Then Return Min
    If Value > Max Then Return Max
    Return Value
End Function

Function CapInt%( Value% , Min%,Max% )
    If Value < Min Then Return Min
    If Value > Max Then Return Max
    Return Value
End Function


; Comparison

Function HighestFloat#( Value1# , Value2# )
    If Value1 > Value2
        Return Value1
    Else 
        Return Value2
    EndIf 
End Function

Function HighestInt%( Value1% , Value2% )
    If Value1 > Value2
        Return Value1
    Else 
        Return Value2
    EndIf 
End Function

Function LowestFloat#( Value1# , Value2# )
    If Value1 < Value2
        Return Value1
    Else 
        Return Value2
    EndIf 
End Function

Function LowestInt%( Value1% , Value2% )
    If Value1 < Value2
        Return Value1
    Else 
        Return Value2
    EndIf 
End Function


; Step limiting (interleaving)

Function InterleaveInt%( Angle% , RotationInterval%=5,Angles%=360 )
    Return Int( Float( RangeInt( Angle,Angles ) - RotationInterval ) / RotationInterval ) * RotationInterval + RotationInterval
End Function


; Angles

; return relative angle difference (takes angle wrapping into account)
; - max angle difference: 359 degrees
Function AngleDifference( CurrentAngle , TargetAngle )
    If TargetAngle - CurrentAngle > 180 Then TargetAngle = TargetAngle - 360
    If CurrentAngle - TargetAngle > 180 Then CurrentAngle = CurrentAngle - 360
    Return CurrentAngle - TargetAngle
End Function

; calculate intelligent hit angle with constant object and projectile speed
Const Math_InterceptAngle_OutOfRange = -1001
Function InterceptAngle( ex#,ey# , evx#,evy# , px#,py# , pvx#,pvy# , bs# , maxtime=0 )
    ;---------------------------------------------------------------------------;-----------------------------------
    ;   Find projectile intercept angle from Source (Enemy) to Target (Player)  ; Algorithm: Eelco <eelco_hoogendoorn@hotmail.com>
    ;---------------------------------------------------------------------------; Wrapper: Teasy <thechange@yahoo.com>
    ;   Parameters                                                              ; Source: INTER~29.BB
    ;---------------------------------------------------------------------------------------------------------------
    ;   (ex,ey)     Enemy position
    ;   evx & evy   Enemy velocity vector
    ;   (px,py)     Player position
    ;   pvx & pvy   Player velocity vector
    ;   bs          Bullet speed
    ;   maxradius   Radius for the player to be in (now or future), 0 = no limit
    ;   maxtime     Only shoot if (future) time of impact is below this amount
    ;               (0 = no limit)
    ;---------------------------------------------------------------------------
;
    ;---------------------------------------------------------------------------
    ;   Make player motion relative to enemy motion
    ;---------------------------------------------------------------------------
;
    es# = Sqr( evx*evx + evy*evy )
    ea# = ATan2( evy,evx )
;
    ps# = Sqr( pvx*pvx + pvy*pvy )
    pa# = ATan2( pvy,pvx )
;
    rpvx#=pvx-evx                                                               ; Relative player velocity
    rpvy#=pvy-evy                                                               ; (Cart vector)
;
    ps=Sqr( rpvx*rpvx + rpvy*rpvy )                                             ; Relative player motion (polar) vector
    pa=ATan2( -rpvy,rpvx )                                                      ; Warning: overrides ps,pa
;
    ;---------------------------------------------------------------------------
    ;   Relative distance to player from enemy
    ;---------------------------------------------------------------------------
;
    rx#=px-ex                                                                   ; Relative player-enemy distance
    ry#=py-ey                                                                   ; (Cart vector)
;
    ;---------------------------------------------------------------------------
    ;   Player must be moving
    ;---------------------------------------------------------------------------
;
    If Abs( ps )>=0.000001                                                      ; If player speed is more than a little
;   :
        ;-----------------------------------------------------------------------
        ;   Find where player will be, using player motion and bullet speed
        ;-----------------------------------------------------------------------
;   :
        a#=ATan2( ey-py,ex-px )                                                 ; Angle player->enemy
        n#=Sin( -pa-a )/bs*ps                                                   ; Angle relative to player, speed scaled
;   :
        ;-----------------------------------------------------------------------
        ;   Player path must be within range for bullet to hit
        ;-----------------------------------------------------------------------
;   :
        If n>=-1 And n<=1                                                       ; Can bullet ever hit player?
;   :   ;
            ;-------------------------------------------------------------------
            ;   New bullet motion for impact
            ;-------------------------------------------------------------------
;   :   ;
            ba#=180+ASin( n )-a                                                 ; Rotated angle back to enemy
;   :   ;
            bvx#=Cos( ba )*bs                                                   ; New bullet velocity (cart) vector
            bvy#=-Sin( ba )*bs
;   :   ;
            ;-------------------------------------------------------------------
            ;   See when the bullet will impact
            ;-------------------------------------------------------------------
;   :   ;
            d#=bvy*pvx - bvx*pvy                                                ; Intersection
;   :   ;
            If Abs( d )>=0.000001                                               ; Avoid infinitely small
                t#=( bvx*ry - bvy*rx )/d                                        ; Time of intersection
;   :   ;   |
                If Not ( maxtime>0 And t>=maxtime Or t<0 )                      ; Time must be in future but not too far
                    Return VectorAngle( bvx,bvy )                               ; Convert back to: Up = 0 degrees
                EndIf
;   :   ;   |
            EndIf
;   :   ;
        EndIf
    EndIf
;
    Return Math_InterceptAngle_OutOfRange
End Function


; Collision

Function PointInRect( px,py , rx,ry , w,h )
    If px>rx
        If py>ry
            If px<rx+w-1
                If py<ry+h-1
                    Return True
                EndIf
            EndIf
        EndIf
    EndIf
End Function

