;-------------------------------------------------------------------------------
; Timer System Version 4 - Special Edition (no global MilliSecs)
; Created by Teasy <TheChange@yahoo.com>
; Last updated: 29 Jun 2007
;-------------------------------------------------------------------------------
; Algorithm:
;
;   This system uses time-hopping which is very accurate but also
;   very math-intensive.  Very similar to buffered, unsychronized
;   transfers, time-hopping makes a variable leap nearest to the
;   current time after a timer's duration.  This without having
;   to regularly update/poll, plus knowing the number of times
;   a timer's time has elapsed.
;-------------------------------------------------------------------------------
; Usage:
;
;   handle = NewTimer( duration [, paused ] )
;       Create a timer, either paused (true) or running (false, default)
;
;   TimerDone( handle )
;      Updates timer and returns the number of elapses
;
;   PauseTimer( handle )
;       Pause, remember time
;
;   ResumeTimer( handle )
;       Unpause, restore time
;
;   StopTimer( handle )
;       Disable timer
;
;   StartTimer( handle )
;       Activate timer and run
;
;   ResetTimer( handle [, duration [, paused ] ] )
;      Restart timer from current point in time
;      Optionally change duration (0 = no change)
;      Changes pause state to False by default
;
;   SetTimer( handle , duration )
;       Change duration, without resetting the current time
;
;   ElapseTimer( handle )
;       Force elapse of timer at current point in time
;
;   FreeTimer( handle )
;       Delete timer, free from memory
;
;-------------------------------------------------------------------------------
; Time units: All durations and such are in milliseconds.
;-------------------------------------------------------------------------------
; Example1:  ; How the system works
;            timer = NewTimer( 1000 )
;            Print "Timer is set to 1 second (1000 ms)"
;            Print "Hit <Spacebar> to check timer"
;            Repeat
;                If Keyhit( 57 )  ; Space
;                    updates = TimerDone( timer )
;                    Write "Number of timer updates"
;                    Print " since last spacebar: "+updates
;                End If
;            Until Keyhit( 1 )  ; Esc
;            End
;-------------------------------------------------------------------------------
; Example2:  ; Blinking with FOR
;            timer = NewTimer( 25 )  ; Interval = 25 ms
;            SetBuffer BackBuffer()
;            Repeat
;                updates = TimerDone( timer )
;                For update = 1 To updates
;                    visible = Not visible
;                Next
;                If visible
;                    Color 255,255,255
;                Else
;                    Color 0,0,0
;                EndIf
;                Text 0,0 , "[..BLINK..]"
;                Flip
;                Cls
;            Until KeyHit( 1 )  ; Esc
;            End
;-------------------------------------------------------------------------------
; Example3:  ; Blinking with MOD
;            timer = NewTimer( 1 )  ; Interval = 1 ms !
;            SetBuffer BackBuffer()
;            Repeat
;                updates = TimerDone( timer )
;                visible = ( visible + updates ) Mod 2
;                If visible
;                    Color 255,255,255
;                Else
;                    Color 0,0,0
;                EndIf
;                Text 0,0 , "[..BLINK..]"
;                Flip False
;                Cls
;            Until KeyHit( 1 )  ; Esc
;            End
;-------------------------------------------------------------------------------
; Known limitations: 32-bit integers allow pausing up to less than 25 days.
;-------------------------------------------------------------------------------
; (Internal) Low level (No error checking)
Type Timer_Object_
    Field LastUpdate
    Field Duration
    Field Paused,Active
    Field Difference
End Type
Global Timer_Object__Count
Function Create_Timer_Object.Timer_Object_( Started , Duration , Paused )
    p.Timer_Object_ = New Timer_Object_
    p\LastUpdate = Started
    p\Duration = Duration
    p\Paused = Paused
    p\Active = True
    p\Difference = 0
    Timer_Object__Count = Timer_Object__Count + 1
    Return p
End Function
Function Delete_Timer_Object( p.Timer_Object_ )
    Delete p
    Timer_Object__Count = Timer_Object__Count - 1
End Function
;-------------------------------------------------------------------------------
; (External) Interface (No error checking)
Function NewTimer( TimeOut , Paused=False )
    Return Handle Create_Timer_Object( MilliSecs() , TimeOut , Paused )
End Function
Function TimerDone( Timer_Handle )
    p.Timer_Object_ = Object.Timer_Object_( Timer_Handle )
    ; Timer must not be paused and must be running
    If Not p\Paused And p\Active
        ; How much time has passed
        Difference = MilliSecs() - p\LastUpdate
        ; At least one update must have passed
        If Difference >= p\Duration
            ; How many updates in time passed
            Updates = Difference / p\Duration
            ; Set time ahead number of updates done
            p\LastUpdate = MilliSecs() - ( Difference - Updates * p\Duration )
            ; Return number of times timer's time has elapsed
            Return Updates
        EndIf
    EndIf
End Function
Function ResetTimer( Timer_Handle , TimeOut=0 , Paused=False )
    p.Timer_Object_ = Object.Timer_Object_( Timer_Handle )
    ; Sync time to now
    p\LastUpdate = MilliSecs()
    ; If change in duration specified
    If TimeOut <> 0
        ; Change duration
        p\Duration = TimeOut
    EndIf
    ; Change pause state (and forget time difference after pause)
    p\Paused = Paused
    p\Difference = 0
End Function
Function SetTimer( Timer_Handle , TimeOut )
    p.Timer_Object_ = Object.Timer_Object_( Timer_Handle )
    ; Change duration
    p\Duration = TimeOut
End Function
Function ElapseTimer( Timer_Handle )
    p.Timer_Object_ = Object.Timer_Object_( Timer_Handle )
    ; Force elapse of timer at current point in time
    p\LastUpdate = MilliSecs() - p\Duration
End Function
Function PauseTimer( Timer_Handle )
    p.Timer_Object_ = Object.Timer_Object_( Timer_Handle )
    ; Can't pause a stopped timer (must be active)
    If p\Active
        p\Paused = True
        ; Remember difference between LastUpdate and MilliSecs (for unpause)
        p\Difference = MilliSecs() - p\LastUpdate
    EndIf
End Function
Function ResumeTimer( Timer_Handle )
    p.Timer_Object_ = Object.Timer_Object_( Timer_Handle )
    ; Can't unpause a stopped timer (must be active)
    If p\Active
        p\Paused = False
        ; Difference is restored: Timer is set 'Difference' behind MilliSecs
        p\LastUpdate = MilliSecs() - p\Difference
    EndIf
End Function
Function StopTimer( Timer_Handle )
    p.Timer_Object_ = Object.Timer_Object_( Timer_Handle )
    ; Reset paused (and time difference for unpause)
    p\Paused = False
    p\Difference = 0
    ; Deactivate timer
    p\Active = False
End Function
Function StartTimer( Timer_Handle )
    p.Timer_Object_ = Object.Timer_Object_( Timer_Handle )
    ; Reset paused (and time difference for unpause)
    p\Paused = False
    p\Difference = 0
    ; Activate timer
    p\Active = True
End Function
Function DeleteTimer( Timer_Handle )
    Delete_Timer_Object Object.Timer_Object_( Timer_Handle )
End Function
